"""
lexers.py - lexer for the rebnf language

Copyright (C) 2023 opsocket <opsocket@pm.me>

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from ply import lex


class Token(lex.LexToken):
    """
    This class describes a simple token wrapper.
    """

    def __init__(self, token):
        assert isinstance(token, lex.LexToken)
        super().__init__()
        self.__dict__.update(token.__dict__)

    def __str__(self):
        return super().__str__()[3:]
      

class REBNFLexer(object):
    tokens = (
        "REGEX",
        "IDENTIFIER",
        "ASSIGN",
        "GROUP_BEGIN",
        "GROUP_END",
        "OPTIONAL_BEGIN",
        "OPTIONAL_END",
        "REPETITION_BEGIN",
        "REPETITION_END",
        "SINGLE_QUOTE_STRING",
        "DOUBLE_QUOTE_STRING",
        "TERMINATOR",
        "OPTIONAL_OP",
        "ONE_OR_MORE_OP",
        "ZERO_OR_MORE_OP",
        "EXCEPT",
        "CONCAT",
        "ALTERNATE",
        "COMMENT_SINGLE",
        "COMMENT_MULTI",
    )

    t_ASSIGN = r"=|:=|::="
    t_GROUP_BEGIN = r"\("
    t_GROUP_END = r"\)"
    t_OPTIONAL_BEGIN = r"\["
    t_OPTIONAL_END = r"\]"
    t_REPETITION_BEGIN = r"{"
    t_REPETITION_END = r"}"
    t_TERMINATOR = r";|\."
    t_OPTIONAL_OP = r"\?"
    t_ONE_OR_MORE_OP = r"\+"
    t_ZERO_OR_MORE_OP = r"\*"
    t_EXCEPT = r"-"
    t_CONCAT = r","
    t_ALTERNATE = r"\|"
    t_COMMENT_SINGLE = r"\#.*"
    t_ignore = " \t\r\f\b"

    def t_REGEX(self, t):
        r"r(\'(\\.|[^\\\'])*\'|\"(\\.|[^\\\"])*\")"
        return t

    def t_IDENTIFIER(self, t):
        r"<[a-zA-Z][a-zA-Z0-9_]*>|[a-zA-Z][a-zA-Z0-9_]*"
        # support for optional identifier enclosures
        if t.value.startswith("<") and t.value.endswith(">"):
            t.value = t.value[1:-1]
        return t

    def t_COMMENT_MULTI(self, t):
        r"\(\*(?:.|\n)*\*\)"
        t.lexer.lineno += t.value.count("\n")
        return t

    def t_SINGLE_QUOTE_STRING(self, t):
        r"\'(\\.|[^\\\'])*\'"
        t.value = t.value[1:-1]
        return t

    def t_DOUBLE_QUOTE_STRING(self, t):
        r"\"(\\.|[^\\\"])*\" "
        t.value = t.value[1:-1]
        return t

    def t_NEWLINE(self, t):
        r"\n+"
        t.lexer.lineno += t.value.count("\n")

    def t_SYMBOL(self, t):
        r"<|>|-|\n|\t|\r|\f|\b"
        return t

    def t_error(self, t):
        print(f"\033[91millegal character: {t.value[0]}\033[0m")
        t.lexer.skip(1)

    def __init__(self):
        # build lexer
        self.lexer = lex.lex(module=self)

    def token(self):
        return self.lexer.token()

    def tokenize(self, code):
        self.lexer.input(code)
        output = []
        while token := self.lexer.token():
            tok = Token(token)
            output.append(tok)
        return output



"""
parsers.py - parser for the rebnf language

Copyright (C) 2023 opsocket <opsocket@pm.me>

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from ply import yacc
from . import lexers


class REBNFParser(object):
    """
    This class describes a rebnf parser.
    """

    tokens = lexers.REBNFLexer.tokens

    def __init__(self):
        self.lexer = lexers.REBNFLexer()
        self.parser = yacc.yacc(module=self)

    def p_grammar(self, p):
        """
        grammar : comment grammar
                | rule grammar
                | comment
                | rule
        """
        print(list(p))
        # if len(self, p) == 3:
        #     p[0] = ('GRAMMAR', p[1])
        # elif len(self, p) == 2:
        #     print('here', list(self, p))
        #     p[0] = p[1]

    def p_comment(self, p):
        """comment : COMMENT_SINGLE
        | COMMENT_MULTI
        """
        p[0] = ("COMMENT", p[1])

    def p_rule(self, p):
        """rule : lhs ASSIGN rhs TERMINATOR"""
        p[0] = ("RULE", (p[1], p[3]))

    def p_lhs(self, p):
        """lhs : IDENTIFIER"""
        p[0] = p[1]

    def p_rhs(self, p):
        """rhs : alternation"""
        p[0] = p[1]

    def p_alternation(self, p):
        """
        alternation : concatenation ALTERNATE alternation
                    | concatenation
        """
        if len(self, p) == 4:
            p[0] = ("ALTERNATION", (p[1], p[3]))
        else:
            p[0] = p[1]

    def p_concatenation(self, p):
        """
        concatenation : factor CONCAT concatenation
                      | factor
        """
        if len(self, p) == 4:
            p[0] = ("CONCAT", (p[1], p[2]))
        else:
            p[0] = p[1]

    def p_factor(self, p):
        """
        factor : term OPTIONAL_OP
               | term ZERO_OR_MORE_OP
               | term ONE_OR_MORE_OP
               | term EXCEPT term
               | term
        """
        if len(self, p) == 3:
            if p[2] == "?*+":
                p[0] = (p[2], p[1])
        elif len(self, p) == 4:
            p[0] = ("EXCEPT", (p[1], p[2]))
        elif len(self, p) == 2:
            p[0] = p[1]

    def p_term(self, p):
        """
        term : GROUP_BEGIN rhs GROUP_END
             | OPTIONAL_BEGIN rhs OPTIONAL_END
             | REPETITION_BEGIN rhs REPETITION_END
             | regex
             | terminal
             | identifier
        """
        if len(self, p) == 4:
            if p[1] == "(":
                p[0] = ("GROUP", p[2])
            if p[1] == "[":
                p[0] = ("OPTIONAL", p[2])
            if p[1] == "{":
                p[0] = ("REPETITION", p[2])
        elif len(self, p) == 2:
            p[0] = p[1]

    def p_term_regex(self, p):
        """regex : REGEX"""
        p[0] = ("REGEX", p[1])

    def p_terminal(self, p):
        """
        terminal : single_quote_string
                 | double_quote_string
        """
        p[0] = ("STRING", p[1])

    def p_single_quote_string(self, p):
        """single_quote_string : SINGLE_QUOTE_STRING"""
        p[0] = ("SINGLE", p[1])

    def p_double_quote_string(self, p):
        """double_quote_string : DOUBLE_QUOTE_STRING"""
        p[0] = ("DOUBLE", p[1])

    def p_identifier(self, p):
        """identifier : IDENTIFIER"""
        p[0] = ("IDENTIFIER", p[1])

    def find_column(self, input, token):
        line_start = input.rfind("\n", 0, token.lexpos) + 1
        return (token.lexpos - line_start) + 1

    def p_error(self, t):
        if t is not None:
            print(
                f"syntax error at line {t.lineno}, column {self.find_column(t.lexer.lexdata, t)}: unexpected token '{t.value}'"
            )
        return t



"""
__main__.py - command-line interface

Copyright (C) 2023 opsocket <opsocket@pm.me>

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from . import tokenize, __ascii__, __version__
import argparse

__description__ = Rf"""

{__ascii__} v{__version__}

"""

def main():
    parser = argparse.ArgumentParser("tokenize", description=__description__, formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument(
        "filepath", type=str, help="filepath to read for tokenization"
    )
    parser.add_argument("-p", "--pretty", action='store_true', help="prints a pretty output")

    args = parser.parse_args()

    if args.filepath:
        with open(args.filepath, "r") as code:
            tokens = tokenize(code.read())
        if args.pretty:
            from pprint import pprint
            pprint(tokens)
        else:
            print(tokens)


if __name__ == "__main__":
    main()
